package observation;

import java.rmi.*;
import java.rmi.server.*;
import java.net.*;
import java.util.*;

/**
 * This class is the coordinator service. Only one coordinator is used
 * into a network.
 * It is the one started from the machine indicated by the string
 * Info#coordMachine.
 * @see Info#coordMachine
 */
public class Coord extends UnicastRemoteObject implements CoordI
{
    private LogI log;
    private Hashtable srcTable, imgTable;
    private Source srcSrcTable, srcImgTable;

/***********************************************************************/

    /**
     * Adds a source. Called by a source.
     * @param src the source remote object
     * @param sName the name of the source
     * @exception ObsException if a source with the same name exists
     */
    public void addSource(SourceI src, String sName)
                throws RemoteException, ObsException{
        log.writeEntry("Coord : addSource " + sName + " : " + src);
        // ajoute l'entree dans la table des sources
        SrcItem srcInfo = new SrcItem(src);
        if (srcTable.containsKey(sName) == true)
            throw new ObsException(ObsException.duplicateSrc + sName);
        srcTable.put(sName, srcInfo);
        if (! sName.equals(Info.srcName) && ! sName.equals(Info.imgName))
            srcSrcTable.changed(srcTable); // non pour les deux sources du coord
    }

/***********************************************************************/

    /**
     * Remove a source. Called by a source.
     * @param sName the name of the source.
     * @exception ObsException if the source doesn't exist.
     */
    public void removeSource(String sName)
                throws RemoteException, ObsException{
        log.writeEntry("Coord : removeSource " + sName);
        // informe toutes ses images et efface les entrees corresp. des images
        Enumeration e = imgTable.keys();   // enumeration des cles des images
        while (e.hasMoreElements()){
            Object k = e.nextElement();    // obtient la cle
            ImgItem ii = (ImgItem)imgTable.get(k);
            if (sName.equals(ii.srcName)){
                // avertit l'image
                ii.image.removeSource();
                // efface l'entree du tableau des images
                imgTable.remove(k);
                srcImgTable.changed(imgTable);
            }
        }
        
        // verifie si la source existe dans la table des sources
        if (! srcTable.containsKey(sName))
            throw new ObsException(ObsException.unregisteredSrc + sName);
        // efface l'entree de la source
        srcTable.remove(sName);
        srcSrcTable.changed(srcTable);
    }

/***********************************************************************/

    /**
     * Remove an image. Called by an image.
     * @param iName the image name.
     * @exception ObsException if the image doesn't exist.
     */
    public void removeImage(String iName)
                throws RemoteException, ObsException{
        log.writeEntry("Coord : removeImage " + iName);
        // verifie si l'image existe dans la table des images
        if (! imgTable.containsKey(iName))
            throw new ObsException(ObsException.unregisteredImg + iName);
        // obtient l'item de l'image iName
        ImgItem imgInfo = (ImgItem)imgTable.get(iName);
        // informe la source:
        SrcItem srcInfo = (SrcItem)srcTable.get(imgInfo.srcName);
        srcInfo.source.removeImage(iName);
        // efface l'entree locale:
        imgTable.remove(iName);
        srcImgTable.changed(imgTable);
    }

/***********************************************************************/

    /**
     * Adds an image. Called by an image.
     * @param img the image remote object.
     * @param iName the image name.
     * @param sName the name of the observed source.
     * @param cond the condition :  0 to update when the source change,
     *     n to update every n ms.
     * @exception ObsException if an image with the same name exists.
     * @return true if ok, false if the source doesn't exist. The caller may
     *     call a while after.
     */
    public boolean addImage(ImageI img, String iName, String sName, int cond)
                throws RemoteException, ObsException{
//        log.writeEntry("Coord : addImage " + iName + " : " + img +
//                       " : " + sName + " : " + cond);
        // verifie si une image avec le meme nom est deja enregistree
        ImgItem imgInfo = new ImgItem(img, sName);
        if (imgTable.containsKey(iName) == true)
            throw new ObsException(ObsException.duplicateImg + iName);
          // il faut mettre aussi cond dans la table des images ?
        // informe la source associee
        SrcItem sI = (SrcItem)srcTable.get(sName);
        if (sI == null)      // source non trouvee => return false
            return false;
//            throw new ObsException(ObsException.unregisteredSrc + sName);
        log.writeEntry("Coord : addImage " + iName + " : " + img +
                       " : " + sName + " : " + cond);
        // ajoute l'image dans la table des images
        imgTable.put(iName, imgInfo);
        srcImgTable.changed(imgTable);
        // informe la source
        sI.source.addImage(img, iName, cond);
        return true;
    }

/***********************************************************************/

    /**
     * Returns the list of all the sources.
     */
    public String [] getSources() throws RemoteException{
        log.writeEntry("Coord : getSources");
        return getTable(srcTable);
    }

/***********************************************************************/

    /**
     * Returns the list of all the images.
     */
    public String [] getImages() throws RemoteException{
        log.writeEntry("Coord : getImages");
        return getTable(imgTable);
    }

/***********************************************************************/

    /**
     * Called by getSources and getImages methods. Returns a String of the
     * table.
     */
    private String [] getTable(Hashtable table){
        String [] ret = new String [table.size()];  // alloue la matrice
        Enumeration e = table.keys();  // cree une enumeration
        int i = 0;
        while (e.hasMoreElements())
            ret[i++] = (String)(e.nextElement());
        return ret;
    }

/***********************************************************************/

    /**
     * Returns the value of the source with the specified name.
     * @param sName the name of the source
     * @exception ObsException if the source doesn't exist.
     */
    public Object getSourceValue(String sName)
                  throws RemoteException, ObsException{
        log.writeEntry("Coord : getSourceValue " + sName);
        if (! srcTable.containsKey(sName))
            throw new ObsException(ObsException.unregisteredSrc + sName);
        return ((SrcItem)srcTable.get(sName)).source.getValue();
    }

/***********************************************************************/

    /**
     * Creates the coordinator.
     */
    public Coord() throws Exception{
        if (log == null)
            log = Info.lookupLog();
        srcTable = new Hashtable();
        imgTable = new Hashtable();
    }

    // cree un objet distant service
    public static void main(String args[]){
        try{
            Info.setRMISecurityManager();
            Coord c = new Coord();
            Naming.bind("rmi://" + InetAddress.getLocalHost().getHostName() +
                        "/" + Info.coordName, c);
            // ces deux lignes ne peuvent pas etre dans le constructeur car
            // elles appellent Coord.addSource() :
            c.srcSrcTable = new Source(c.srcTable, Info.srcName);
            c.srcImgTable = new Source(c.imgTable, Info.imgName);
            System.out.println("Service coord binded");
        } catch(Exception e){
            e.printStackTrace();
        }
    }
}

