package observation;

import java.rmi.*;
import java.rmi.server.*;
import java.net.*;
 
/**********************************************************************/

/**
 * This thread waits t milliseconds and then finishes. It is used like
 * temporisation for addImage method.
 */
class Wait extends Thread{
    int time;
    public Wait(int t){
        time = t;
    }
    public void run(){
        try{
            Thread.sleep(time);
        }catch(InterruptedException e){
            e.printStackTrace();   // apparait si un autre thread le reveille
        }
    }
}

/**********************************************************************/

public class Image extends UnicastRemoteObject implements ImageI
{
    private static CoordI coord;
    private static LogI log;
    private String varName;   // le nom externe de la variable
    public Object var;      // la variable (publique)

/**********************************************************************/

    /**
     * Updates an image. Called directly by its source.
     * @param o the new value of the source.
     */
    public void update(Object o) throws RemoteException{
        log.writeEntry("Image " + varName + " : update : " + o.toString());
        var = o;      // COPIE, pas changement de handle
//        System.out.println(varName + " : " + o.toString());
    }

/**********************************************************************/

    /**
     * Informs the image that its source will be removed. Called by the
     * coordinator.
     */
    public void removeSource() throws RemoteException{
        log.writeEntry("Image " + varName + " : removeSource");
        // ex.: en fct. d'un param. plusieurs politiques existent :
        // - throws exception
        // - gele la valeur
    }

/**********************************************************************/

    /**
     * Removes the image. Called by the application which holds the image.
     * This method calls the removeImage of the Coordinator, which
     * informs the appropriate source about the end of the image.
     * After remove, the image can't be used anymore and will be
     * garbage-collected.
     */
    public void remove() throws RemoteException{
        // faut-il mettre un flag a 1 signifiant la terminaison de l'image ?
        log.writeEntry("Image " + varName + " : remove");
        try{
            coord.removeImage(varName);   // pas d'exception normalement
        }catch(ObsException oe){
            oe.printStackTrace();
        }
    }

/**********************************************************************/

    /**
     * Returns the value of a source. Called by any process. Note that
     * this method is static (it doesn't need to have an image).
     * @param sName the name of the source.
     * @return the value of the source.
     * @exception ObsException if a source with the name sName has not
     *     yet registered.
     */
    static public Object getSourceValue(String sName) throws Exception{
        if (coord == null)
            coord = Info.lookupCoord();
        return coord.getSourceValue(sName);
    }

/**********************************************************************/

    /**
     * Creates an image and initializes it with the source value. This
     * method will return either delay milliseconds, or if an exception is
     * thrown. This method also register it at coordinator. Every
     * image-variable must have an Image object associated with it.
     * @param iName the name of the image.
     * @param sName the name of the source to observe.
     * @param cond the condition. If 0 or less than 0, the image will
     *     be updated every time the source changes. If a value greater
     *     than 0, the image will be updated every cond milliseconds.
     *     Because of the fact that the updates are made by threads, even
     *     for cond=0, some values may be lost (if the source change faster
     *     than the OS process-switch), but the image will always have the
     *     most recently value of the source.
     * @param delay the time to wait for the source to register
     *     at coordinator. The image tries every 500ms to see if the source is
     *     registered. After at least delay milliseconds, the image
     *     throws an ObsException.
     * @exception ObsException if the source is not registered after
     *     delay milliseconds or if an image with the same name has
     *     already registered at the coordinator.
     */
    public Image(String iName, String sName, int cond, int delay)
                 throws Exception{
        Info.setRMISecurityManager();
        if (log == null)
            log = Info.lookupLog();
        if (coord == null)
            coord = Info.lookupCoord();
        varName = iName;
        if (cond < 0)
            cond = 0;
        if (delay <= 0)
            delay = 0;  // or 1 ?
        if (! coord.addImage(this, varName, sName, cond)){ //not found=>attente
            Wait w = new Wait(delay);
            boolean srcFound = false;
            w.start();
            while(w.isAlive()){
                Thread.sleep(500);    // ou Thread.yield()
                if (coord.addImage(this, varName, sName, cond)){    // retry
                    // tue le thread (car il n'est plus necessaire)
                    w.stop();
                    srcFound = true;
                    break;
                }
            }
            if (! srcFound)
                throw new ObsException(ObsException.unregisteredSrc + sName);
        }
        // initialise l'image avec la valeur de la source
        var = getSourceValue(sName);
          /*
          On pouvait utiliser la valeur de retour de Coord.addImage, mais il
          fallait avoir un moyen de savoir si la source est enregistree ou non
          (la source pouvait avoir aussi la valeur null).
          */
        log.writeEntry("Image " + iName + " : created " + " : " + sName +
                       " : " + cond + " : " + delay);
    }
}

