package observation;

import java.rmi.*;
import java.rmi.server.*;
import java.net.*;
import java.util.*;

public class Source extends UnicastRemoteObject implements SourceI
{
    private static CoordI coord;
    private static LogI log;
    private Object var;    // la variable source
    private String varName;   // le nom externe de la variable source
    private Hashtable imgTable;   // le vecteur des images

/***********************************************************************/

    /** A source table element. */
    class ImgElem{
        ImageI img;
        ImageHandler imgHand;   // thread gerant la communication
        ImgElem(ImageI i, ImageHandler ih){
            img = i;
            imgHand = ih;
        }
    }

/***********************************************************************/

// condition: si 0, rafraichissement au changement de la source
//            si n, rafraichissement toutes les n ms

    /**
     * The thread associated to the source-image link. It updates the
     * image.
     */
    class ImageHandler extends Thread{
        private ImageI img;
        private int condition;
        private boolean end = false;
        ImageHandler(ImageI i, int c) throws RemoteException{
            img = i;
            condition = c;
        }
        void finish(){
            end = true;
            if (condition != 0)
                interrupt();
            else
                resume();
        }
        void wakeup(){
            if (condition == 0)
                resume();
        }
        public void run(){
            try{
                if (condition != 0)    // rafraichissement periodique
                    while(! end){
                        // l'image est d'abord rafraichie (ou apres sleep)
                        synchronized(var){
                            img.update(var);
                        }
                        sleep(condition);
                    }
                else    // rafraichissement au changement
                    while(! end){
                        synchronized(var){
                            img.update(var);
                        }
                        suspend();
                    }
            }catch(Exception e){
                e.printStackTrace();
            }
            try{
                img.update(var);   // le dernier rafraichissement
            }catch(Exception e){
                e.printStackTrace();
            }
        }
    }

/***********************************************************************/

    /**
     * Adds an image. Called by the coordinator.
     * @param i the image reference.
     * @param iName the image name.
     * @param cond the condition.
     * @see Coord#addImage
     */
    public void addImage(ImageI i, String iName, int cond)
                         throws RemoteException{
        log.writeEntry("Source " + varName + " : addImage " + iName +
                       " : " + cond + " : " + i);
        // cree le thread associe a l'image
        ImageHandler ih = new ImageHandler(i, cond);
        ih.start();
        // ajoute l'image dans la table locale des images
        ImgElem imgInfo = new ImgElem(i, ih);
        imgTable.put(iName, imgInfo);
    }

/***********************************************************************/

    /**
     * Removes an image. Called by the coordinator.
     * @param iName the name of the image to be removed.
     */
    public void removeImage(String name) throws RemoteException{
        log.writeEntry("Source " + varName + " : removeImage " + name);
        // tue le thread associe a l'image
        ((ImgElem)imgTable.get(name)).imgHand.stop();
        // efface l'entree locale du tableau des images
        imgTable.remove(name);
    }

/***********************************************************************/

    /**
     * Removes the source. Called by the application which holds the source.
     * This method calls the removeSource of the Coordinator, which
     * informs all the images about the end of the source. Before removing,
     * it sends its latest value to all its images.
     */
    public void remove() throws RemoteException{
        log.writeEntry("Source " + varName + " : remove");
        // informe tous les threads de finir
        Enumeration e = imgTable.elements();
        while (e.hasMoreElements())
            ((ImgElem)e.nextElement()).imgHand.finish();  // pb 1 de synchro ?
        // attend le dernier rafraichissement et la fin des threads
        e = imgTable.elements();
        while (e.hasMoreElements())
            try{
                ((ImgElem)e.nextElement()).imgHand.join();  // pb 2 de synchro ?
            }catch(Exception oe){
                oe.printStackTrace();
            }
        imgTable.clear();    // inutile, mais sur (utile pour gc ?)
        // informe le coord
        try{
            coord.removeSource(varName);   // pas d'exception normalement
        }catch(ObsException oe){
            oe.printStackTrace();
        }
    }

/***********************************************************************/

    /** Returns the value of the source. Called by the coordinator. */
    public Object getValue() throws RemoteException{
        return var;
    }

/***********************************************************************/

    /**
     * Informs this object that its source has changed. Called by the
     * program which holds the source every time it changes the source
     * value. This class use the SAME source-variable, not a copy.
     * This means that if the program changes the source value and
     * the changed method is not called, the images will receive the
     * new value, but the update condition will not work correctly.
     * @param o the new value of the object.
     */
    public void changed(Object o){
        // met a jour la variable:
        var = o;
        // reveille tous les threads images (ou seulm. ceux qui sont bloques) :
        Enumeration e = imgTable.elements();
        while(e.hasMoreElements()){
            ImgElem ii = (ImgElem)(e.nextElement());
            ii.imgHand.wakeup();
        }
    }

/***********************************************************************/

    /**
     * Creates a source. This method also registers it at the coordinator.
     * Every source-variable must have a Source object associated with it.
     * Any source must implement the Serialization interface.
     * @param sVar the source variable (the initial value of the source).
     * @param sName the name of the source.
     * @exception ObsException if a source with the same name has already
     *     registered at the coordinator.
     */
    public Source(Object sVar, String sName) throws Exception{
        Info.setRMISecurityManager();
        if (log == null)
            log = Info.lookupLog();
        if (coord == null)
            coord = Info.lookupCoord();
        var = sVar;
        varName = sName;
        imgTable = new Hashtable();
        coord.addSource(this, varName);
        log.writeEntry("Source " + sName + " : created ");
    }
}

